from .basicCommands import BasicCommands
from .dataCommands import DataCommands
from cryptography.fernet import Fernet
import json
import os
class MineDB(BasicCommands, DataCommands):

    def __init__(self):
        self.version = "1.0.3"
        self.existing_db={"sample":{"data":{"version":"1.1v","developer":"hrs_developers"}}}
        self.currDB = "sample"
        self.currColl = "data"

        base_dir = os.path.dirname(os.path.abspath(__file__))
        self.__path = os.path.join(base_dir, "secure.dat")
        self.__key_path = os.path.join(base_dir, "MineDBKey.key")

        os.makedirs(base_dir, exist_ok=True)

        #checking for key
        #if present load
        if os.path.exists(self.__key_path):
            with open(self.__key_path, "rb") as f:
                key = f.read()
        #if not present create
        else:
            key = Fernet.generate_key()
            with open(self.__key_path, "wb") as f:
                f.write(key)

        self.__cipher = Fernet(key)

        #checking for exsting_db
        #if present load
        if os.path.exists(self.__path):
            with open(self.__path, "rb") as f:
                encrypted_data = f.read()
            decrypted_data = self.__cipher.decrypt(encrypted_data)
            self.existing_db = json.loads(decrypted_data)
        #if not present create with sample
        else:
            self.existing_db = {
                "sample": {
                    "data": {
                        "version": self.version,
                        "developer": "hrs_developers"
                    }
                }
            }
            self.save()    

    def save(self):
        json_data = json.dumps(self.existing_db).encode()
        encrypted_data = self.__cipher.encrypt(json_data)
        with open(self.__path,"wb") as f:
            f.write(encrypted_data)



class DataCommands:

#datatype and value validation commands

    #checks function attribute should be string
    def __checkIns(self, value, source=None):
        for v in value:
            if v == None:
                continue
            if not isinstance(v, str):
                raise ValueError(f"Error : valueError : {source} : Value should be string")
        return  True

    #checks for loading/modifing value datatype
    def __checkValueType(self, value, source=None):
        if isinstance(value,str):
            if len(value) > 1:
                return "text"
            else:
                return "chr"
        elif isinstance(value,bool):
            return 'bool'
        elif isinstance(value,int):
            return 'int'
        elif isinstance(value, float):
            return 'float'
        else:
            raise ValueError(f"Error : {source} : Invalid data type of value {value}")

    #checking valiadation of loading/modifing data
    def __checkValue(self, loadData, source=None):
        for key, value in loadData.items():
            if key in self.existing_db[self.currDB][self.currColl]:
                if value == None or value == "None":
                    continue
                if self.__checkValueType(value,source) != self.existing_db[self.currDB][self.currColl][key]["dataType"]:
                    raise ValueError(f"Error : {source} : Invalid data type of value {value}")
            else:
                raise KeyError(f"Error : {source} : Field not exist")
        return True

#get and set commands
    def getDB(self):
        return self.currDB

    def getCollection(self):
        return self.currColl

    def setDB(self, dbName):
        if not self.__checkIns([dbName],"setDB"):
            return
        if dbName.isidentifier():
            if dbName in self.existing_db:
                self.currDB = dbName
            else:
                raise KeyError("Error : setDB : Database not exist")
        else:
            raise ValueError("Error : setDB : Invalid identifier")

    def setCollection(self, colName):
        if not self.__checkIns([colName],"setCollection"):
            return
        if colName.isidentifier():
            if colName in self.existing_db[self.currDB]:
                self.currColl = colName
            else:
                raise KeyError("Error : setCollection : Collection not exist")
        else:
            raise ValueError("Error : setCollection : Invalid identifier")

#major data commands
    def load(self, **kwargs):
        if len(self.existing_db[self.currDB][self.currColl]) == len(kwargs):
            if not self.__checkValue(kwargs,"load"):
                return
            for key, value in kwargs.items():
                self.existing_db[self.currDB][self.currColl][key]['items'].append(value)
            return True
        else:
            raise ValueError("Error : load : Incomplete loading")

    def modify(self, search_by, search_value, update_field, new_value):
        collection = self.existing_db[self.currDB][self.currColl]

        # validate fields and value types
        self.__checkValue(
            {search_by: search_value, update_field: new_value},
            "modify"
        )

        value_array = collection[search_by]["items"]

        # find matching row indexes
        indexes = [
            i for i, v in enumerate(value_array)
            if v == search_value
        ]

        if not indexes:
            raise ValueError("modify: search_value not exist")

        # update all matching rows
        for i in indexes:
            collection[update_field]["items"][i] = new_value

        return True

    def remove(self, search_by, search_value):
        collection = self.existing_db[self.currDB][self.currColl]

        # validate search field and value type
        self.__checkValue({search_by: search_value}, "remove")

        value_array = collection[search_by]["items"]

        # find all matching indexes
        indexes = [
            i for i, v in enumerate(value_array)
            if v == search_value
        ]

        if not indexes:
            raise ValueError("remove: search_value not exist")

        # remove from highest index to lowest (no drift)
        for i in reversed(indexes):
            for field in collection:
                collection[field]["items"].pop(i)

        return True

    def explore(self,*args):
        exp = {}
        for value in args:
            if value in self.existing_db[self.currDB][self.currColl]:
                exp[value]=self.existing_db[self.currDB][self.currColl][value]["items"]
            else:
                raise KeyError("Error : explore : Field not exist")
        return exp
    
    def exploreAll(self):
        try:
            exp={}
            for value in self.existing_db[self.currDB][self.currColl]:
                exp[value]=self.existing_db[self.currDB][self.currColl][value]["items"]
            return exp
        except(TypeError, KeyError) as e:
            raise RuntimeError("Error : exploreAll : something went wrong, check setup of database and column") from e


import pytest
from MineDB.basicCommands import BasicCommands


class DummyDB(BasicCommands):
    def __init__(self):
        self.existing_db = {}


@pytest.fixture
def db():
    return DummyDB()


def test_create_database(db):
    db.createDB("testdb")
    assert "testdb" in db.existing_db


def test_create_collection(db):
    db.createDB("testdb")
    db.createCollection(
        "testdb",
        "users",
        id="int",
        active="bool"
    )
    assert "users" in db.existing_db["testdb"]


def test_duplicate_database_raises(db):
    db.createDB("testdb")
    with pytest.raises(ValueError):
        db.createDB("testdb")


def test_duplicate_collection_raises(db):
    db.createDB("testdb")
    db.createCollection("testdb", "users", id="int")
    with pytest.raises(ValueError):
        db.createCollection("testdb", "users", id="int")


def test_alter_add_field(db):
    db.createDB("testdb")
    db.createCollection("testdb", "users", id="int")

    db.alterAddField("testdb", "users", "active", "bool")
    assert "active" in db.existing_db["testdb"]["users"]


def test_alter_drop_field(db):
    db.createDB("testdb")
    db.createCollection("testdb", "users", id="int", active="bool")

    db.alterDropField("testdb", "users", "active")
    assert "active" not in db.existing_db["testdb"]["users"]


def test_alter_field_type_conversion(db):
    db.createDB("testdb")
    db.createCollection("testdb", "users", age="int")

    db.existing_db["testdb"]["users"]["age"]["items"] = [1, 2, 3]

    db.alterFieldType("testdb", "users", "age", "float")

    assert db.existing_db["testdb"]["users"]["age"]["items"] == [1.0, 2.0, 3.0]


def test_rename_database(db):
    db.createDB("old")
    db.renameDB("old", "new")

    assert "new" in db.existing_db
    assert "old" not in db.existing_db


def test_rename_collection(db):
    db.createDB("testdb")
    db.createCollection("testdb", "users", id="int")

    db.renameCollection("testdb", "users", "accounts")
    assert "accounts" in db.existing_db["testdb"]



class BasicCommands:

    #checks function attribute should be string
    def __checkIns(self, value=None, source=None):
        if value == None:
            value = []
        for v in value:
            if v == None:
                continue
            if not isinstance(v, str):
                raise ValueError(f"Error : valueError : {source} : Value should be string")
        return  True

    #checks attribute items lenght to prevent corruption
    def __checkItemLength(self, dbName, colName):
        if dbName not in self.existing_db:
            return False
        if colName not in self.existing_db[dbName]:
            return False

        collection = self.existing_db[dbName][colName]

        # empty collection is valid
        if not collection:
            return 0

        lengths = {
            len(field["items"])
            for field in collection.values()
        }

        if len(lengths) != 1:
            return False

        return lengths.pop()

#database commands

    def createDB(self,dbName:str):
        if not self.__checkIns([dbName], "createDB"):
            return
        checkName = dbName.isidentifier()
        if checkName == False:
            raise ValueError("Error : createDB : Invalid identifier")
        elif dbName in self.existing_db:
            raise ValueError("Error : createDB : Database is already exists")
        else:
            self.existing_db[dbName]={}
            return True

    def showDB(self,dbName:str=None):
        if not self.__checkIns([dbName], "showDB"):
            return
        if dbName != None:
            if dbName in self.existing_db:
                return True
            else:
                raise KeyError("Error : showDB : Database not exist")
        else:
            existingDB = []
            for db in self.existing_db.keys(): 
                existingDB.append(db)
            return existingDB
    
    def dropDB(self,dbName:str):
        if dbName == "sample":
            raise ValueError("MineDB : sample database is not dropable")
        if not self.__checkIns([dbName], "dropDB"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : dropDB : Database not exist")
        self.existing_db.pop(dbName)
    
    def renameDB(self, dbName:str, newName:str):
        if not self.__checkIns([dbName, newName], "renameDB"):
            return
        if dbName == "sample":
            raise ValueError("MineDB : sample database is not renameable")
        tempDB = {}

        if dbName not in self.existing_db:
            raise KeyError("Error : renameDB : Database not exist")
        if newName in self.existing_db:
            raise ValueError("Error : renameDB : Database already exist")
        checkName = newName.isidentifier()
        if checkName != True:
            raise ValueError("Error : renameDB : Invalid identifier")
        for key, value in self.existing_db.items():
            if key == dbName:
                tempDB[newName] = value
            else:
                tempDB[key]=value
        self.existing_db = tempDB
        return True

#collection commands

    def createCollection(self,dbName:str,colName:str,**kwargs):
        if not self.__checkIns([dbName, colName], "createCollection"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName in  self.existing_db[dbName]:
            raise ValueError("Error : alterFieldType : Collection already exist")
        checkName = colName.isidentifier()
        if checkName != True:
            raise ValueError("Error : createCollection : Invalid Identifier Of Attribute")
        temp = {}
        for key, value in kwargs.items():
            if key.isidentifier():
                if(value in ["text", "int", "float", "bool", "chr"]):
                    temp[key]={"dataType" : value, "items" : []}
                else:
                    raise ValueError(f"Error : createCollection : invalid data type it should be : {["text", "int", "float", "bool", "chr"]}")
            else:
                raise ValueError("Error : createCollection : Invalid Identifier Of Attribute")
        self.existing_db[dbName][colName]=temp
        return True

    def showCollection(self,dbName:str,colName=None):
        if not self.__checkIns([dbName,colName],"showCollection"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName != None:
            if colName not in  self.existing_db[dbName]:
                raise KeyError("Error : alterFieldType : Collection not exist")
            collection = {colName : {}}
            for value in self.existing_db[dbName][colName]:
                collection[colName][value]=self.existing_db[dbName][colName][value]['dataType']
            return collection
        else:
            collection = {dbName:[]}
            for col in self.existing_db[dbName]:
                collection[dbName].append(col)
            return collection
    
    def dropCollection(self, dbName:str, colName:str):
        if not self.__checkIns([dbName,colName], "dropCollection"):
            return -1
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName not in  self.existing_db[dbName]:
            raise KeyError("Error : alterFieldType : Collection not exist")
        self.existing_db[dbName].pop(colName)
        return True
    
    def renameCollection(self, dbName:str, colName:str, newName:str):
        if not self.__checkIns([dbName, colName, newName], "renameCollection"):
            return
        tempCol = {}
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName not in  self.existing_db[dbName]:
            raise KeyError("Error : alterFieldType : Collection not exist")
        checkName = newName.isidentifier()
        if checkName != True:
            raise ValueError("Error : alterFieldType : Invalid identifier")
        for key, value in self.existing_db[dbName].items():
            if key == colName:
                tempCol[newName] = value
            else:
                tempCol[key]=value
        self.existing_db[dbName] = tempCol
        return True

 #Alter Commands

    def alterFieldType(self, dbName:str, colName:str, fieldName:str, dataType:str):
        if not self.__checkIns([dbName, colName, fieldName, dataType], "alterFieldType"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName not in  self.existing_db[dbName]:
            raise KeyError("Error : alterFieldType : Collection not exist")
        if fieldName not in self.existing_db[dbName][colName]:
            raise KeyError("Error : alterFieldType : Field not exist")
        if dataType not in ["text", "int", "float", "bool", "chr"]:
            raise ValueError(f"Error : alterFieldType : invalid data type it should be : {["text", "int", "float", "bool", "chr"]}")
        
        field= self.existing_db[dbName][colName][fieldName]
        old_items = self.existing_db[dbName][colName][fieldName]["items"]
        def convert(v):
            if v is None:
                return None
            if dataType == "int":
                return int(v)
            if dataType == "float":
                return float(v)
            if dataType == "bool":
                return bool(v)
            if dataType == "text":
                return str(v)
            if dataType == "chr":
                s = str(v)
                if len(s) != 1:
                    raise ValueError("chr must be length 1")
                return s

        try:
            new_items = [convert(v) for v in old_items]
        except Exception:
            raise ValueError("alterFieldType: incompatible existing data")
        
        field["dataType"] = dataType
        field["items"] = new_items
        return True
    
    def alterFieldName(self, dbName:str, colName:str, fieldName:str, newName:str):
        if not self.__checkIns([dbName,colName,fieldName,newName], "alterFieldName"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterFieldType : Database not exist")
        if colName not in  self.existing_db[dbName]:
            raise KeyError("Error : alterFieldType : Collection not exist")
        if fieldName not in self.existing_db[dbName][colName]:
            raise KeyError("Error : alterFieldType : Field not exist")
        tempFiled = {}
        checkName = newName.isidentifier()
        if checkName != True:
            raise ValueError("Error : alterFieldType : Invalid identifier")
        for key, value in self.existing_db[dbName][colName].items():
            if key == fieldName:
                tempFiled[newName] = value
            else:
                tempFiled[key] = value
        self.existing_db[dbName][colName] = tempFiled 
        return True
    
    def alterDropField(self, dbName:str, colName:str, fieldName:str):
        if not self.__checkIns([dbName, colName, fieldName]):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterDropField : Database not exist")
        if colName not in self.existing_db[dbName]:
            raise KeyError("Error : alterDropField : Collection not exist")
        if fieldName not in self.existing_db[dbName][colName]:
            raise KeyError("Error : alterDropField : Field not exist")
        self.existing_db[dbName][colName].pop(fieldName)
        
        return True

    def alterAddField(self, dbName:str, colName:str, fieldName:str, dataType:str):
        if not self.__checkIns([dbName, colName, fieldName],"alterAddField"):
            return
        if dbName not in self.existing_db:
            raise KeyError("Error : alterAddField : Database not exist")
        if colName not in self.existing_db[dbName]:
            raise KeyError("Error : alterAddField : Collection not exist")
        if fieldName in self.existing_db[dbName][colName]:
            raise ValueError("Error : alterAddField : Field already exist")
        if dataType not in ["text", "int", "float", "bool", "chr"]:
            raise ValueError(f"Error : alterAddField : invalid data type it should be : {["text", "int", "float", "bool", "chr"]}")
        length = self.__checkItemLength(dbName,colName)
        if length is False:
            raise RuntimeError("alterAddField: Internal schema corruption")
        self.existing_db[dbName][colName][fieldName]={"dataType":dataType,"items":[None]*length}
        
        return True


from setuptools import setup, find_packages

setup(
    name="minedb",
    version="1.0.2",
    author="Harsh Singh Sikarwar",
    description="A lightweight encrypted local database using dictionary-based storage",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/harshsikarwar/MineDB",
    project_urls={
        "Source": "https://github.com/harshsikarwar/MineDB",
        "Issues": "https://github.com/harshsikarwar/MineDB/issues",
    },
    license="MIT",
    python_requires=">=3.9",
    install_requires=[
        "cryptography>=41.0.0"
    ],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import pytest
from MineDB.dataCommands import DataCommands
from MineDB.basicCommands import BasicCommands


class DummyDB(BasicCommands, DataCommands):
    def __init__(self):
        self.existing_db = {
            "app": {
                "settings": {
                    "theme": {"dataType": "text", "items": []},
                    "notifications": {"dataType": "bool", "items": []},
                }
            }
        }
        self.currDB = "app"
        self.currColl = "settings"


@pytest.fixture
def db():
    return DummyDB()


def test_load_valid_data(db):
    db.load(theme="dark", notifications=True)

    assert db.existing_db["app"]["settings"]["theme"]["items"] == ["dark"]
    assert db.existing_db["app"]["settings"]["notifications"]["items"] == [True]


def test_load_incomplete_data_raises(db):
    with pytest.raises(ValueError):
        db.load(theme="dark")


def test_load_invalid_type_raises(db):
    with pytest.raises(ValueError):
        db.load(theme=123, notifications=True)


def test_modify_data(db):
    db.load(theme="dark", notifications=True)
    db.modify("theme", "dark", "notifications", False)

    assert db.existing_db["app"]["settings"]["notifications"]["items"] == [False]


def test_modify_missing_value_raises(db):
    db.load(theme="dark", notifications=True)
    with pytest.raises(ValueError):
        db.modify("theme", "light", "notifications", False)


def test_remove_data(db):
    db.load(theme="dark", notifications=True)
    db.remove("theme", "dark")

    assert db.existing_db["app"]["settings"]["theme"]["items"] == []
    assert db.existing_db["app"]["settings"]["notifications"]["items"] == []


def test_remove_missing_value_raises(db):
    db.load(theme="dark", notifications=True)
    with pytest.raises(ValueError):
        db.remove("theme", "light")


def test_no_index_drift(db):
    db.load(theme="dark", notifications=True)
    db.load(theme="light", notifications=False)
    db.remove("theme", "dark")

    assert db.existing_db["app"]["settings"]["theme"]["items"] == ["light"]
    assert db.existing_db["app"]["settings"]["notifications"]["items"] == [False]
